// ---------------------------------------------------------------------------
// BoolTranslator.java
// A class for the translation of various boolean expressions to SMT-LIB strings.
// $Id: BoolTranslator.java,v 1.6 2020/07/21 14:29:14 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import riscal.semantics.Types.Value;
import riscal.smt.ExpressionTranslator.TypeEnumerationInformation;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.AndExp;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.EquivExp;
import riscal.syntax.AST.Expression.ExistsExp;
import riscal.syntax.AST.Expression.FalseExp;
import riscal.syntax.AST.Expression.ForallExp;
import riscal.syntax.AST.Expression.GreaterEqualExp;
import riscal.syntax.AST.Expression.GreaterExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.ImpliesExp;
import riscal.syntax.AST.Expression.InSetExp;
import riscal.syntax.AST.Expression.LessEqualExp;
import riscal.syntax.AST.Expression.LessExp;
import riscal.syntax.AST.Expression.LetExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.NotExp;
import riscal.syntax.AST.Expression.OrExp;
import riscal.syntax.AST.Expression.RecApplicationExp;
import riscal.syntax.AST.Expression.RecIdentifierExp;
import riscal.syntax.AST.Expression.RecordSelectionExp;
import riscal.syntax.AST.Expression.SubsetExp;
import riscal.syntax.AST.Expression.TrueExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.util.Seq;
import riscal.util.Seq.Next;

public class BoolTranslator extends ASTVisitor.Base<String>
{
	/**
	 * Used to translated non boolean expressions.
	 */
	private ExpressionTranslator expTrans;
	
	/**
	 * Determines if skolemisation can be applied.
	 * Note: skolemize==true is not sufficient for an application
	 * of skolemisation. SMT.useSkolemisation must also be true.
	 */
	private boolean skolemize=false;
	
	/**
	 * The default name of the skolem function name.
	 */
	private static final String defaultSkolemFunctionName="f";
	
	/**
	 * The default name of the boolean enumeration function,
	 */
	private static final String defaultTypeEnumerationName="boolEnum";
	
	/**
	 * The SMT-LIB name for the boolean type.
	 */
	private static final String boolTypeString="Bool";
	
	/**
	 * The variables which were already introduced by lets in the current expression.
	 * Note: If the processed expression is not a top level expression (e.g. the parameter of a function)
	 * this would not contain all the necessary variable names. 
	 * But this case is not important as these names are only needed
	 * for skolemisation and skolemisation can only be applied to top level
	 * expressions.
	 */
	private Set<ValueSymbol> letValues=new HashSet<ValueSymbol>();
	
	/***************************************************************************
	 * Create a BoolTranslator
	 * @param et The ExpressionTranslator which shall be used to 
	 * translate non boolean expressions.
	 **************************************************************************/
	public BoolTranslator(ExpressionTranslator et) 
	{
		this.expTrans=et;
	}
	
	/***************************************************************************
	 * Translates a boolean expression to a SMT-LIB string.
	 * @param exp The expression that shall be translated.
	 * @param skolemize If true, skolemisation can be applied.
	 * @return The SMT-LIB representation of the given expression.
	 **************************************************************************/
	public String compute(Expression exp,boolean skolemize)
	{
		boolean oldSkolemize=this.skolemize;
		this.skolemize=skolemize;
		if(skolemize)
		{
			letValues.clear();
		}
		String s=exp.accept(this);
		this.skolemize=oldSkolemize;
		return s;
	}
	
	/***************************************************************************
	 * Computes the SMT-LIB strings for the values of the type.
	 * @param bare If true the values a represented by bitvectors
	 * without the #b prefix.
	 * @return A list containing the SMT-LIB representations of
	 * the boolean values.
	 **************************************************************************/
	public static List<String> computeValues(boolean bare)
	{
		List<String> res=new ArrayList<String>(2);
		if(bare)
		{
			res.add("0");
			res.add("1");
		}
		else
		{
			res.add("false");
			res.add("true");
		}		
		return res;
	}
	
	/***************************************************************************
	 * Computes the SMT-LIB strings for the values of the type.
	 * @param type must be either a BoolType or a SubType with basetype BoolType;
	 * @param seq A sequence containing the semantic values of the type.
	 * @param bare If true the values a represented by bitvectors
	 * without the #b prefix.
	 * @return A list containing all values of the given type represented 
	 * by their SMT-LIB representation.
	 **************************************************************************/
	static List<String> processValueSequence(Type type,Seq<Value> seq,boolean bare)
	{		
		int len=ExpressionTranslator.getTypeSize(type);
		List<String> result=new ArrayList<String>(len);
		Function <Value, String> proc;
		if(bare)
			proc=v -> processValueBare(v);
		else
			proc=v -> processValue(v);
		while (true)
    {
      Next<Value> next = seq.get();
      if (next == null) break;
      	result.add(proc.apply(next.head));
      seq = next.tail;
    }
		return result;
	}
	
	/***************************************************************************
	 * Computes the SMT-LIB representation of the given value.
	 * @param val The value that shall be translated to SMT-LIB
	 * @return The SMT-LIB representation of val.
	 **************************************************************************/
	private static String processValue(Value val)
	{
		boolean boolVal=((riscal.semantics.Types.Value.Bool) val).getValue();
		return Boolean.toString(boolVal);
	}
	
	/***************************************************************************
	 * Computes the SMT-LIB representation of the given value.
	 * In contrast to processValue the result does not
	 * contain a leading "#b"
	 * @param val The value that shall be translated to SMT-LIB
	 * @return The SMT-LIB representation of val.
	 **************************************************************************/
	private static String processValueBare(Value val)
	{
		boolean boolVal=((riscal.semantics.Types.Value.Bool) val).getValue();
		return boolVal?"1":"0";
	}
	
	/***************************************************************************
	 * Creates an enumeration function for booleans
	 * The values of the type are enumerated according to their
	 * occurrence in riscal.semantics.Values.getValueSeq
	 * @return The name of the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunction()
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		expTrans.addDeclaration("(define-fun "+funName+"((x Bool)) (_ BitVec 1) (ite x #b1 #b0))");		
		return new TypeEnumerationInformation(funName, 1, 2);
	}
	
	/***************************************************************************
	 * Creates an enumeration function for the denoted values.
	 * Defines the function via a table (requires UFs)
	 * The values of the type are enumerated according to their
	 * occurrence in riscal.semantics.Values.getValueSeq
	 * @param lst The list of values that shall be enumerated.
	 * @return The name of the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTable(List<Value> lst)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		expTrans.addDeclaration("(declare-fun "+funName+" (Bool) (_ BitVec 1))");
		long i=0;
		for(Value val:lst)
		{
			expTrans.addDeclaration("(assert (= ("+funName+" "+processValue(val)+") "+IntTranslator.generateSMT_LIB_Constant(i, 1)+"))");
			i++;
		}
		return new TypeEnumerationInformation(funName, 1, lst.size());
	}
	
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTableWithFunctionDefinition(List<Value> lst)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		expTrans.addDeclaration("(define-fun "+funName+" ((x Bool)) (_ BitVec 1) (ite "+processValue(lst.get(0))+" "+
				IntTranslator.generateSMT_LIB_Constant(0, 1)+" "+IntTranslator.generateSMT_LIB_Constant(1, 1)+"))");
		return new TypeEnumerationInformation(funName, 1, lst.size());
	}
	
	/***************************************************************************
	 * Used to translate a boolean value to a SMT-LIB string.
	 **************************************************************************/
	static class BoolValueContainer extends ExpressionTranslator.ValueContainer
	{
		BoolValueContainer(BoolType bT){}
		/***************************************************************************
		 * Translates the denoted value to a SMT-LIB string.
		 **************************************************************************/
		String process(Value val) 
		{
			return processValue(val);
		}
		/***************************************************************************
		 * Translates the denoted value to a bit sequence (in the SMT-LIB style).
		 * The difference to process is that "#b" is not prepended to this sequence.
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			return processValueBare(val);
		}
	}


	/***************************************************************************
	 * Determines the minimal length of a bit vector which is necessary
	 * to represent the values of a BoolType.
	 * @return The vector length that is necessary to represent 
	 * the values from BoolType.
	 **************************************************************************/
	public static long giveVectorLength()
	{
		return 1L;
	}
	
	public String visit(TrueExp exp)
	{
		return "true";
	}
	
	public String visit(FalseExp exp)
	{
		return "false";
	}
	
	
	public String visit(NotExp exp)
	{
		String str=compute(exp.exp, false);
		return "(not "+str+")";
	}
	
	public String visit(AndExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(and "+str1+" "+str2+")";
	}
	
	public String visit(OrExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(or "+str1+" "+str2+")";
	}
	
	public String visit(ImpliesExp exp)
	{
		String str1=compute(exp.exp1,false);
		String str2=exp.exp2.accept(this);
		return "(=> "+str1+" "+str2+")";
	}
	
	public String visit(EquivExp exp)
	{
		String str1=compute(exp.exp1,false);
		String str2=compute(exp.exp2,false);
		return "(= "+str1+" "+str2+")";
	}
	
	public String visit(IfThenElseExp exp)
	{
		String cond=compute(exp.exp1, false);
		String thenExp=exp.exp2.accept(this);
		String elseExp=exp.exp3.accept(this);
		return "(ite "+cond+" "+thenExp+" "+elseExp+")";
	}
	
	public String visit(IdentifierExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(ApplicationExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(RecIdentifierExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(RecApplicationExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(EqualsExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(NotEqualsExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(LessExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(LessEqualExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(GreaterExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(GreaterEqualExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(SubsetExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(InSetExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(Binder b)
	{
		String str=expTrans.compute(b.exp);
		return "( "+b.ident.string+" "+str+")";
	}
	
	public String visit(LetParExp exp)
	{
		for(Binder b:exp.binder)
		{
			letValues.add((ValueSymbol)b.ident.getSymbol());
		}
		String str=Arrays.stream(exp.binder).map(b -> b.accept(this)).collect(Collectors.joining(" "));
		String expStr=exp.exp.accept(this);
		return "(let ("+str+") "+expStr+")";
	}
	
	public String visit(LetExp exp)
	{
		SMT.illegalUsage(exp, "Let expressions should not occur in the translation to SMT-LIB!");
		return null;
	}
	
	
	public String visit(TupleSelectionExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(RecordSelectionExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(MapSelectionExp exp)
	{
		return expTrans.compute(exp);
	}
	

	/***************************************************************************
	 * Computes the free variables in the denoted expression.
	 * @param exp The expression of which we want to compute the free variables.
	 * @return ValueSymbols which correspond to the free variables.
	 **************************************************************************/
	private Set<ValueSymbol> typedFreeSymbolsWithoutValuesAndLets(Expression exp)
	{
		Set<ValueSymbol> freeVariables=expTrans.typedFreeSymbolsWithoutValues(exp);
		freeVariables.removeAll(letValues);
		return freeVariables;
	}
	
	/***************************************************************************
	 * Auxiliary method for the expansion of universal quantifiers.
	 * @param expStr The translation of the quantified expression.
	 * @param variable The name of the quantified variable.
	 * @param vals The SMT-LIB representation of the values of the quantified type.
	 * @return Expansion of the corresponding quantifier.
	 **************************************************************************/
	private String expandForall(String expStr,String variable,List<String> vals)
	{
		if(vals.size()==0)
			return "true";
		String letStr=vals.stream().map(str -> "(let (("+variable+" "+str+")) "+expStr+")").collect(Collectors.joining(" "));
		if(vals.size()==1)//boolector does not support and with 1 argument (Z3 does)
			return letStr;
		else
			return "(and "+letStr+")";
	}
	
	/***************************************************************************
	 * Computes a conjunction of the elements of the list
	 * @param lst A list whose elements represent SMT-LIB boolean expressions 
	 * @return A SMT-LIB conjunction of the elements of the given list
	 **************************************************************************/
	static String conjoin(List<String> lst)
	{
		String args=String.join(" ", lst);
		if(lst.size()==1)
		{
			return args;
		}
		return "(and "+args+")";
	}
	
	/***************************************************************************
	 * Applies quantifier expansion on the denoted universal quantifier.
	 * @param exp The quantified expression that shall be expanded.
	 * @return The expansion of the quantifier given in the SMT-LIB language.
	 **************************************************************************/
	private String expandForall(ForallExp exp)
	{
		String variable=exp.qvar.qvcore[0].ident.string;
		Type ty=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
		List<String> vals=expTrans.computeValues(ty);
		if(vals.size()==0)
		{
			return Boolean.toString(true);
		}	
		String aux=exp.exp.accept(this);
		if(ty instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
		{
			String cond=expTrans.generateTypeConditions(ty, variable);
			if(cond!=null)
			{
				aux="(=> "+cond+" "+aux+")";
			}				
		}
		String expStr=aux;
		/*
		 * This is necessary because (nested) quantifiers could lead to Strings which are so 
		 * long that the java string functionalities come to their limit.
		 */
		if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), expStr.length()))
		{
			Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp.exp);
			//Really necessary - shouldn't the wexp be null?
//			Expression we=exp.qvar.qvcore[0].wexp;
//			if(we!=null)
//			{
//				freeSyms.addAll(expTrans.typedFreeSymbolsWithoutValues(we));
//			}
			if(freeSyms.contains(exp.qvar.qvcore[0].ident.getSymbol()))
			{
				if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
				{
					List<String> lst=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms, 
							(ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol(), boolTypeString);
					return conjoin(lst);
				}
				else
				{
					expStr=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms,boolTypeString);
				}
			}
			/*
			 * If this is the case the quantified expression does not depend on the quantified variable.
			 * Thus we do not need to construct the conjunction over all values of the quantified type.
			 * Note: We apply this check only in this branch of the code as here we have the set freeSyms anyways.
			 * Thus it is cheap to check this.
			 * Otherwise this set is not available. We assume most of the time the quantified expression
			 * will depend on the quantified variable. Thus probably most of the time the computation
			 * of this set would we useless, thus we do not do it.
			 */
			else
			{
				return expStr;
			}		
		}
		return expandForall(expStr, variable, vals);
	}
	
	/***************************************************************************
	 * Constructs an universally quantified SMT-LIB string.
	 * @param expStr The translation of the quantified expression. 
	 * @param variable The name of the quantified variable.
	 * @param ty The quantified type.
	 * @return An universally quantified SMT-LIB expression.
	 **************************************************************************/
	public String constructSMTLibForall(String expStr,String variable,Type ty)
	{
		String condition=expTrans.generateTypeConditions(ty, variable);
		String typeString=expTrans.translateType(ty);
		String SMTLIbQVar="(("+variable+" "+typeString+"))";
		if(condition==null)
		{
			return "(forall "+SMTLIbQVar+" "+expStr+")";
		}
		else
		{
			return "(forall "+SMTLIbQVar+" (=> "+condition+" "+expStr+"))";
		}
	}
	
	/***************************************************************************
	 * Auxiliary method for the expansion of universal quantifiers.
	 * @param expStr The translation of the quantified expression. 
	 * @param variable The name of the quantified variable.
	 * @param ty The quantified type.
	 * @return An universally quantified SMT-LIB expression.
	 **************************************************************************/
	public String constructForallExpansion(String expStr,String variable,Type ty)
	{
		//Conditions should not be necessary here.
//		String condition=expTrans.generateTypeConditions(ty, variable);
//		if(condition!=null)
//		{
//			expStr="(=> "+condition+" "+expStr+")";
//		}
		if(SMT.expandBaseTypesOfSubTypes&&ty instanceof SubType)
		{
			String condition=expTrans.generateTypeConditions(ty, variable);
			if(condition!=null)
			{
				expStr="(=> "+condition+" "+expStr+")";
			}
		}
		List<String> vals=expTrans.computeValues(ty);
		return expandForall(expStr, variable, vals);
	}
	
	
	public String visit(ForallExp exp)
	{
		//Since we assume that the Preprocessor was applied to the expression we only have to deal with IdentifierTypeQuantifiedVar
		//and we can assume that there is only one QuantifiedVariableCore.
		assert exp.qvar.qvcore.length==1;
		if(expTrans.getUnquantify())
		{
			return expandForall(exp);
		}
		else
		{
			Type t=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
			return constructSMTLibForall(exp.exp.accept(this), exp.qvar.qvcore[0].ident.string, t);
		}			
	}
	
	/***************************************************************************
	 * Auxiliary method for the expansion of existential quantifiers.
	 * @param expStr The translation of the quantified expression.
	 * @param variable The name of the quantified variable.
	 * @param vals The SMT-LIB string of the values of the quantified type.
	 * @return Expansion of the corresponding quantifier.
	 **************************************************************************/
	private String expandExists(String expStr,String variable,List<String> vals)
	{
		if(vals.size()==0)
			return "false";
		String letStr=vals.stream().map(str -> "(let (("+variable+" "+str+")) "+expStr+")").collect(Collectors.joining(" "));
		if(vals.size()==1)//boolector does not support and with 1 argument (Z3 does)
			return letStr;
		else
			return "(or "+letStr+")";
	}
	
	/***************************************************************************
	 * Computes a disjunction of the elements of the list
	 * @param lst A list whose elements represent SMT-LIB boolean expressions 
	 * @return A SMT-LIB disjunction of the elements of the given list
	 **************************************************************************/
	static String disjoin(List<String> lst)
	{
		String args=String.join(" ", lst);
		if(lst.size()==1)
		{
			return args;
		}
		return "(or "+args+")";
	}
	
	/***************************************************************************
	 * Applies quantifier expansion on the denoted existential quantifier.
	 * @param exp The quantified expression that shall be expanded.
	 * @return The expansion of the quantifier given in the SMT-LIB language.
	 **************************************************************************/
	private String expandExists(ExistsExp exp)
	{
		String variable=exp.qvar.qvcore[0].ident.string;
		Type ty=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
		List<String> vals=expTrans.computeValues(ty);
		if(vals.size()==0)
		{
			return Boolean.toString(false);
		}	
		String aux=exp.exp.accept(this);
		if(ty instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
		{
			String cond=expTrans.generateTypeConditions(ty, variable);
			if(cond!=null)
			{
				aux="(and "+cond+" "+aux+")";
			}
		}
		String expStr=aux;
		/*
		 * This is necessary because (nested) quantifiers could lead to Strings which are so 
		 * long that the java string functionalities come to their limit.
		 */
		if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), expStr.length()))
		{
			Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp.exp);
			//Really necessary - shouldn't the wexp be null?
//			Expression we=exp.qvar.qvcore[0].wexp;
//			if(we!=null)
//			{
//				freeSyms.addAll(expTrans.typedFreeSymbolsWithoutValues(we));
//			}
			if(freeSyms.contains(exp.qvar.qvcore[0].ident.getSymbol()))
			{
				if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
				{
					List<String> lst=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms, 
							(ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol(), boolTypeString);
					return disjoin(lst);
				}
				else
				{
					expStr=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms,boolTypeString);
				}
			}
			/*
			 * If this is the case the quantified expression does not depend on the quantified variable.
			 * Thus we do not need to construct the disjunction over all values of the quantified type.
			 * Note: We apply this check only in this branch of the code as here we have the set freeSyms anyways.
			 * Thus it is cheap to check this.
			 * Otherwise this set is not available. We assume most of the time the quantified expression
			 * will depend on the quantified variable. Thus probably most of the time the computation
			 * of this set would we useless, thus we do not do it.
			 */
			else
			{
				return expStr;
			}		
		}
		return expandExists(expStr, variable, vals);
	}
	
	/***************************************************************************
	 * If conditions are needed in order to ensure that a bit vector corresponds to
	 * a value from a existentially quantified RISCAL type then we have to
	 * ensure that a skolem function fulfills this condition. As we cannot
	 * define a skolem function we have to ensure the condition
	 * by universally quantifying over the parameters of the skolem function.
	 * These universal quantifiers we have to remove by expansion.
	 * Thus it is possible that the usage of a skolem function requires
	 * a larger expansion than the expansion of the quantifier itself.
	 * This method shall heuristically determine whether skolemisation
	 * pays off or not.
	 * @param quantifiedType The quantified type.
	 * @param dependingSymbols The symbols that correspond to the required parameters
	 * of the skolem function.
	 * @return true if skolemisation shall be applied, otherwise false.
	 **************************************************************************/
	private boolean skolemisationHeuristic(Type quantifiedType,Set<ValueSymbol> dependingSymbols)
	{
		if(SMT.getUseSkolemisation()==SMT.SkolemisationMode.NeverSkolemise)
			return false;
		if(SMT.getUseSkolemisation()==SMT.SkolemisationMode.AlwaysSkolemise)
			return true;
		if(!expTrans.hasTypeCondition(quantifiedType))
			return true;
		long typeSize=ExpressionTranslator.getTypeSize(quantifiedType);
		long valuesSize=1;
		for(ValueSymbol sym:dependingSymbols)
		{
			valuesSize+=ExpressionTranslator.getTypeSize(sym.type);
		}
		if(valuesSize>typeSize)
			return false;
		else
			return true;
		
	}
	
	/***************************************************************************
	 * Applies skolemisation to the denoted existential quantifier,
	 * @param exp The expression that shall be skolemised.
	 * @return The skolemised translation of the expression.
	 **************************************************************************/
	private String skolemise(ExistsExp exp)
	{
		Set<ValueSymbol> freeVariables=typedFreeSymbolsWithoutValuesAndLets(exp);
		List<ValueSymbol> sortedFreeVars=Utilities.sortSymbols(freeVariables);
		Type variableType=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
		if(!skolemisationHeuristic(variableType, freeVariables))
			return expandExists(exp);
		String skolemFunctionName=expTrans.getFunctionName(defaultSkolemFunctionName);
		if(skolemFunctionName.equals(exp.qvar.qvcore[0].ident.string))
			skolemFunctionName=expTrans.getFunctionName(defaultSkolemFunctionName);
		//The symbols collected should be ValueSymbols.		
		String paramTypes=sortedFreeVars.stream().map(sym -> expTrans.translateType(
				Utilities.getBaseType(((ValueSymbol)sym).type))).collect(Collectors.joining(" "));		
		String skolFunctionDeclaration="(declare-fun "+skolemFunctionName+" ("+paramTypes+")"+
				expTrans.translateType(Utilities.getBaseType(variableType))+")";
		expTrans.addDeclarationRemovable(skolFunctionDeclaration);
		if(!(variableType instanceof BoolType)&&
				(!(variableType instanceof SetType)||SMT.expandBaseTypesOfSubTypes))
		{
			List<String> names=new ArrayList<String>();
			List<Type>types=new ArrayList<Type>();
			for(ValueSymbol sym:sortedFreeVars)
			{
				names.add(sym.ident.string);
				types.add(sym.type);
			}
			String cond=expTrans.generateTypeConditions(variableType, 
					expTrans.generateSMTLIBFunctionApplication(skolemFunctionName, names));
			if(cond!=null)
			{
				expTrans.addDeclarationRemovable("(assert "+expTrans.generateQuantifiedSMTLibExpression(
						cond, names,types)+")");
			}
		}
		String expStr=exp.exp.accept(this);
		if(freeVariables.isEmpty())
		{
			return "(let (("+exp.qvar.qvcore[0].ident.string+" "+skolemFunctionName+"))"+expStr+")";
		}
		else
		{
			String paramNames=sortedFreeVars.stream().map(sym->sym.ident.string).collect(Collectors.joining(" "));
			return "(let (("+exp.qvar.qvcore[0].ident.string+" ("+skolemFunctionName+" "+paramNames+")))"+expStr+")";
		}
	}
	
	/***************************************************************************
	 * Constructs an existentially quantified SMT-LIB string.
	 * @param expStr The translation of the quantified expression. 
	 * @param variable The name of the quantified variable.
	 * @param ty The quantified type.
	 * @return An existentially quantified SMT-LIB expression.
	 **************************************************************************/
	public String constructSMTLibExists(String expStr,String variable,Type ty)
	{
		String condition=expTrans.generateTypeConditions(ty, variable);
		String typeString=expTrans.translateType(ty);
		String SMTLIbQVar="(("+variable+" "+typeString+"))";
		if(condition==null)
		{
			return "(exists "+SMTLIbQVar+" "+expStr+")";
		}
		else
		{
			return "(exists "+SMTLIbQVar+" (and "+condition+" "+expStr+"))";
		}
	}
	
	public String visit(ExistsExp exp)
	{
		//Since we assume that the Preprocessor was applied to the expression we only have to deal with IdentifierTypeQuantifiedVar
		//and we can assume that there is only one QuantifiedVariableCore.
		assert exp.qvar.qvcore.length==1;		
		if(expTrans.getUnquantify())
		{
			if(SMT.getUseSkolemisation()!=SMT.SkolemisationMode.NeverSkolemise)
			{
				letValues.add((ValueSymbol) exp.qvar.qvcore[0].ident.getSymbol());
				if(skolemize)
					return skolemise(exp);
			}
			return expandExists(exp);
		}
		else
		{
			Type t=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
			return constructSMTLibExists(exp.exp.accept(this),exp.qvar.qvcore[0].ident.string,t);
		}
	}
	
	public String visit(QuantifiedVariable qVar)
	{
		String str=Arrays.stream(qVar.qvcore).map(qv -> qv.accept(this)).collect(Collectors.joining(" "));
		return "("+str+")";
	}
	
	public String visit(IdentifierTypeQuantifiedVar qvcore)
	{
		String typeString=((ValueSymbol)qvcore.ident.getSymbol()).type.accept(expTrans);
		return "("+qvcore.ident.string+" "+typeString+")";
	}
	
	public String visit(IdentifierSetQuantifiedVar qvcore)
	{
		SMT.illegalUsage(qvcore, "An error in the preprocessing occured: IdentifierSetQuantifiedVar cannot be translated to SMT-LIB");
		return null;
	}
	
	public String visit(BoolType bt)
	{
		return boolTypeString;
	}
	
	
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
