// ---------------------------------------------------------------------------
// FreeVariables.java
// Determine the symbols of the free variables of an expression.
// Based on the FreeVariables class
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: TypedFreeVariables.java,v 1.1 2020/02/11 07:22:08 schreine Exp $
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.HashSet;
import java.util.Set;

import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ChooseExp;
import riscal.syntax.AST.Expression.ChooseInElseExp;
import riscal.syntax.AST.Expression.ChooseInExp;
import riscal.syntax.AST.Expression.ExistsExp;
import riscal.syntax.AST.Expression.ForallExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.LetExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MatchExp;
import riscal.syntax.AST.Expression.MaxExp;
import riscal.syntax.AST.Expression.MinExp;
import riscal.syntax.AST.Expression.NumberExp;
import riscal.syntax.AST.Expression.ProductExp;
import riscal.syntax.AST.Expression.SetBuilderExp;
import riscal.syntax.AST.Expression.SumExp;
import riscal.syntax.AST.Parameter;
import riscal.syntax.AST.PatternExpression;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol.ValueSymbol;

public class TypedFreeVariables extends ASTVisitor.Base<Void>
{
	//the set of free variables computed
	//should be a set of valuesymbols.
	private Set<ValueSymbol> free = new HashSet<ValueSymbol>();
	
  /***************************************************************************
   * Compute the free variables of an expression.
   * @param exp the expression.
   * @return its free variables (including constants).
   **************************************************************************/
  public static Set<ValueSymbol> compute(Expression exp)
  {
  	TypedFreeVariables visitor = new TypedFreeVariables();
    exp.accept(visitor);
    return visitor.free;
  }
    
  /***************************************************************************
   * Compute the free variables of a quantified variable core
   * @param qvar the quantified variable core.
   * @return its free variables (including constants).
   **************************************************************************/
  public static Set<ValueSymbol> compute(QuantifiedVariableCore qvc)
  {
  	Set<ValueSymbol> free = new HashSet<ValueSymbol>();
    if (qvc.wexp != null) free.addAll(compute(qvc.wexp));
    free.remove(qvc.ident.getSymbol());
    if (qvc instanceof IdentifierSetQuantifiedVar)
    {
      IdentifierSetQuantifiedVar qvc0 = (IdentifierSetQuantifiedVar)qvc; 
      free.addAll(compute(qvc0.exp));
    }
    else if (qvc instanceof IdentifierTypeQuantifiedVar)
    {
      IdentifierTypeQuantifiedVar qvc0 = (IdentifierTypeQuantifiedVar)qvc; 
      free.addAll(compute(qvc0.type));
    }
    else 
      assert false;
    return free;
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Add free variables of quantifier
   * @param qvar the quantified variable
   * @param exp the body expression (possibly null)
   **************************************************************************/
  private void addFree(QuantifiedVariable qvar, Expression exp)
  {
    Set<ValueSymbol> free0 = exp == null ? new HashSet<ValueSymbol>() : compute(exp);
    if (qvar.exp != null) free0.addAll(compute(qvar.exp));
    int n = qvar.qvcore.length;
    for (int i = n-1; i >= 0; i--)
    {
      QuantifiedVariableCore qvc = qvar.qvcore[i];
      free0.remove(qvc.ident.getSymbol());
      free0.addAll(compute(qvc));
    }
    free.addAll(free0);
  }
  
  
  /***************************************************************************
   * Compute the free variables of a type
   * @param type the type
   * @return its free variables (including constants).
   **************************************************************************/
  private static Set<ValueSymbol> compute(Type type)
  {
  	TypedFreeVariables visitor = new TypedFreeVariables();
    type.accept(visitor);
    return visitor.free;
  }
  
  // --------------------------------------------------------------------------
  //
  // the visitor methods (adding free variables by their traversal)
  //
  // --------------------------------------------------------------------------
  
  public Void visit(IdentifierExp exp)
  {
    free.add((ValueSymbol)exp.ident.getSymbol());
    return null;
  }
  
  public Void visit(ChooseExp exp)
  {
    addFree(exp.qvar, null);
    return null;
  }
  
  public Void visit(ChooseInElseExp exp)
  {
    addFree(exp.qvar, exp.exp1);
    exp.exp2.accept(this);
    return null;
  }
  
  public Void visit(ChooseInExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(ExistsExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(ForallExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(LetExp exp)
  {
    Set<ValueSymbol> free0 = compute(exp.exp);
    int n = exp.binder.length;
    for (int i = n-1; i >= 0; i--)
    {
      Binder binder = exp.binder[i];
      free0.remove(binder.ident.getSymbol());
      free0.addAll(compute(binder.exp));
    }
    free.addAll(free0);
    return null;
  }
  
  public Void visit(LetParExp exp)
  {
    Set<ValueSymbol> free0 = compute(exp.exp);
    for (Binder binder : exp.binder)
      free0.remove(binder.ident.getSymbol());
    for (Binder binder : exp.binder)
      free0.addAll(compute(binder.exp));
    free.addAll(free0);
    return null;
  }
  
  public Void visit(MatchExp exp)
  {
    exp.exp.accept(this);
    for (PatternExpression pexp : exp.pexp)
    {
      Set<ValueSymbol> pfree = compute(pexp.exp);
      if (pexp instanceof PatternExpression.ApplicationPatternExp)
      {
        PatternExpression.ApplicationPatternExp pexp0 =
            (PatternExpression.ApplicationPatternExp)pexp;
        for (Parameter p : pexp0.params)
        {
          pfree.remove(p.ident.getSymbol());
          pfree.addAll(compute(p.type));
        }
      }
      free.addAll(pfree);
    }
    return null;
  }
  
  public Void visit(MaxExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(MinExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(NumberExp exp)
  {
    addFree(exp.qvar, null);
    return null;
  }
  
  public Void visit(ProductExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(SetBuilderExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
  
  public Void visit(SumExp exp)
  {
    addFree(exp.qvar, exp.exp);
    return null;
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------