// ---------------------------------------------------------------------------
// TaskTree.java
// A tree representing checking tasks.
// $Id: TaskTree.java,v 1.21 2020/05/25 13:10:44 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2017-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.swt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.function.Consumer;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import riscal.Main;
import riscal.syntax.AST.SourcePosition;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.*;
import riscal.tasks.*;
import riscal.types.Environment.Symbol.*;
import riscal.smt.*;

public class TaskTree
{
  // the SWT tree (subclassing not allowed)
  private Tree tree;
  
  // the top window
  private final TopWindow top;
  
  // the task colors
  private Color taskColorOpen; 
  private Color taskColorClosed;
  
  // the color for non-executable tasks that are also not folders
  private Color noTaskColor;
  
  // the function to which the tree currently belongs (null, if none)
  private FunctionSymbol fun;
  
  /***************************************************************************
   * Create a tree displaying task structures.
   * @param parent the parent widget.
   * @param top the top window.
   **************************************************************************/
  public TaskTree(Composite parent, TopWindow top)
  {    
    tree = new Tree(parent, SWT.SINGLE);
    this.top = top;
    createMenu();
    createHandlers();
    taskColorOpen = tree.getDisplay().getSystemColor(SWT.COLOR_RED);
    taskColorClosed = tree.getDisplay().getSystemColor(SWT.COLOR_BLUE);
    noTaskColor = tree.getDisplay().getSystemColor(SWT.COLOR_DARK_MAGENTA);
  }
  
  /****************************************************************************
   * Get visual representation of tree.
   * @return the tree.
   ***************************************************************************/
  public Tree getTree()
  {
    return tree;
  }
  
  /***************************************************************************
   * Create menu for tree
   **************************************************************************/
  private void createMenu()
  {
    final Menu menu = new Menu(tree.getShell(), SWT.POP_UP);
    tree.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        for (int i=0; i<menuItems.length; i++) {
          menuItems[i].dispose ();
        }
        TreeItem [] treeItems = tree.getSelection();
        if (treeItems.length != 1) return;
        final TreeItem item = treeItems[0];
        Task task = (Task)item.getData();
        if (task.getFunction() != null)
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Execute Task");
          menuItem.setImage(top.getStartImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              executeTask(task, item);
            }
          });
        }
        if (task.getRefutation() != null)
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Show Counterexample");
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              executeRefutation(task, item);
            }
          });
        }
        if (task instanceof TaskFolder)
        {
          TaskFolder folder = (TaskFolder)task;
          MenuItem menuItem = new MenuItem(menu, SWT.PUSH);
          menuItem.setText("Execute All Tasks");
          menuItem.setImage(top.getStartImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              executeSubtasks(folder);
            }
          });
          if (SMT.enabled)
          {
            MenuItem menuItem0 = new MenuItem(menu, SWT.PUSH);
            menuItem0.setText("Apply SMT Solver to All Theorems");
            menuItem0.setImage(null);
            menuItem0.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                applySMT(folder);
              }
            });
          }
        }
        final String description = task.getDescription();
        if (description != null)       
        {
          MenuItem menuItem = new MenuItem(menu, SWT.PUSH);
          menuItem.setText("Print Description");
          menuItem.setImage(null);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              Main.println(description);
            }
          });
        }
        final String definition = task.getDefinition();
        if (definition != null)       
        {
          MenuItem menuItem = new MenuItem(menu, SWT.PUSH);
          menuItem.setText("Print Definition");
          menuItem.setImage(null);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              Main.println(definition);
            }
          });
          if (SMT.enabled)
          {
            // apply SMT solver to check validity of theorem in given context
            FunctionSymbol fun = task.getFunction();
            List<Declaration> declarations = task.getDeclarations();
            if (fun != null && declarations != null)
            {
              Declaration def = fun.getDefinition();
              if (def instanceof TheoremParamDefinition)
              {
                TheoremParamDefinition theorem = (TheoremParamDefinition)def;
                MenuItem menuItem0 = new MenuItem(menu, SWT.PUSH);
                menuItem0.setText("Apply SMT Solver");
                menuItem0.setImage(null);
                menuItem0.addSelectionListener(new SelectionAdapter() {
                  public void widgetSelected(SelectionEvent e) {
                    applySMT(theorem, declarations, item);
                  }
                });
              }
            }
          }
        }
      };
    });
  }
  
  /****************************************************************************
   * Execute a task associated to the tree item.
   * @param task to be executed.
   * @param the associated tree item.
   ***************************************************************************/
  private void executeTask(Task task, TreeItem item)
  {   
    FunctionSymbol fun = task.getFunction();
    if (fun == null) return;
    top.startExecution(false, new Runnable()
    {
      public void run() 
      {
        boolean okay = MainSWT.execute(fun);
        top.updateView(new Runnable() {
          public void run() {
            tree.deselect(item);
            if (okay)
            {
              item.setForeground(taskColorClosed);
              item.setImage(top.getTaskClosedImage());
            }
          }});
      }
    });
  }
  
  /****************************************************************************
   * Execute refutation associated to the tree item.
   * @param task to be executed.
   * @param the associated tree item.
   ***************************************************************************/
  private void executeRefutation(Task task, TreeItem item)
  {   
    FunctionSymbol fun = task.getRefutation();
    if (fun == null) return;
    top.startExecution(false, new Runnable()
    {
      public void run() 
      {
        MainSWT.execute(fun);
        top.updateView(new Runnable() {
          public void run() {
            tree.deselect(item);
          }});
      }
    });
  }
  
  // the thread to check the validity of a theorem by applying an SMT solver
  // and the result of its application
  private Thread smtThread = null;
  private boolean smtValid = false;

  /****************************************************************************
   * Apply SMT solver to check validity of theorem in the context of declarations
   * @param the theorem to be checked.
   * @param the declarations preceding the theorem.
   * @param item the corresponding menu item
   ***************************************************************************/
  private void applySMT(TheoremParamDefinition theorem,
    List<Declaration> declarations, TreeItem item)
  {   
    top.startExecution(false, new Runnable()
    {
      public void run() 
      {
        applySMTCore(theorem, declarations);
        top.updateView(new Runnable() {
          public void run() {
            tree.deselect(item);
            if (smtValid)
            {
              item.setForeground(taskColorClosed);
              item.setImage(top.getTaskClosedImage());
            }
          }});
      }
    });
  }
  @SuppressWarnings("deprecation")
  private boolean applySMTCore(TheoremParamDefinition theorem,
    List<Declaration> declarations)
  {
    top.run(new Runnable() {
      public void run()
      {
        smtValid = false;
        smtThread = new Thread(()-> 
        { 
          try
          {
            smtValid = SMT.check(theorem, declarations); 
          }
          catch(Main.Stopped e) 
          { 
            Main.getOutput().println("ERROR: execution was stopped.");
          }
        });
        smtThread.start();
        try
        {
          while (true)
          {
            try { Thread.sleep(100); } catch(InterruptedException e) { }
            if (!smtThread.isAlive()) break;
            Main.checkStopped();
          }
        }
        catch(Exception e)
        {
          SMT.terminateSolver();
          smtThread.interrupt();
          try { Thread.sleep(500); } catch(InterruptedException e0) { }
          if (smtThread.isAlive()) 
          {
            Main.getOutput().println("SMT thread could not be gracefully interrupted; it is forcefully stopped.");
            smtThread.stop();
          }
          if (e instanceof Main.Stopped)
            Main.getOutput().println("ERROR: execution was stopped.");
        }
      }});
    return smtValid;
  }
  
  /****************************************************************************
   * Find item associated to task.
   * @param items the items in which the searched item is contained.
   * @param task the task to which the searched item is associated.
   * @return the item (null, if not found).
   ***************************************************************************/
  private TreeItem find(TreeItem[] items, Task task)
  {
    for (TreeItem item : items)
    {
      Task task0 = (Task)item.getData();
      if (task == task0) return item;
      TreeItem result = find(item.getItems(), task);
      if (result != null) return result;
    }
    return null;
  }
  
  /****************************************************************************
   * Synchronously execute a task associated to the tree item.
   * @param task to be executed.
   * @param the associated tree item.
   ***************************************************************************/
  private void executeTaskSync(Task task)
  {   
    FunctionSymbol fun = task.getFunction();
    if (fun == null) return;
    boolean okay = MainSWT.execute(fun);
    top.updateView(new Runnable() {
      public void run() {
        TreeItem item = find(tree.getItems(), task);
        tree.deselect(item);
        if (okay)
        {
          item.setForeground(taskColorClosed);
          item.setImage(top.getTaskClosedImage());
        }
      }});
  };

  /****************************************************************************
   * Synchronously apply SMT solver to the tree item.
   * @param task to be executed.
   * @param the associated tree item.
   ***************************************************************************/
  private void applySMTSync(Task task)
  {   
    FunctionSymbol fun = task.getFunction();
    if (fun == null) return;
    List<Declaration> declarations = task.getDeclarations();
    if (declarations == null) return;
    Declaration def = fun.getDefinition();
    if (!(def instanceof TheoremParamDefinition)) return;
    TheoremParamDefinition theorem = (TheoremParamDefinition)def;
    boolean okay = applySMTCore(theorem, declarations);
    top.updateView(new Runnable() {
      public void run() {
        TreeItem item = find(tree.getItems(), task);
        tree.deselect(item);
        if (okay)
        {
          item.setForeground(taskColorClosed);
          item.setImage(top.getTaskClosedImage());
        }
      }});
  }

  /****************************************************************************
   * Synchronously execute all subtasks recursively.
   * @param the folder containing the tasks (and other folders).
   ***************************************************************************/
  private void executeSubtasksSync(TaskFolder folder)
  {
    Main.resetStopped();
    for (Task task: folder.getTasks())
    {
      executeTaskSync(task);
      if (Main.wasStopped()) return;
    }
    for (TaskFolder folder0: folder.getFolders())
    {
      executeSubtasksSync(folder0);
      if (Main.wasStopped()) return;
    }
  }

  /****************************************************************************
   * Synchronously apply SMT solver recursively.
   * @param the folder containing the tasks (and other folders).
   ***************************************************************************/
  private void applySMTSync(TaskFolder folder)
  {
    Main.resetStopped();
    for (Task task: folder.getTasks())
    {
      applySMTSync(task);
      if (Main.wasStopped()) return;
    }
    for (TaskFolder folder0: folder.getFolders())
    {
      applySMTSync(folder0);
      if (Main.wasStopped()) return;
    }
  }
  
  /****************************************************************************
   * Recursively execute all subtasks.
   * @param the folder containing the tasks (and other folders).
   ***************************************************************************/
  private void executeSubtasks(TaskFolder folder)
  {
    top.startExecution(false, new Runnable()
    {
      public void run() 
      {
        executeSubtasksSync(folder);
      }
    });
  }

  /****************************************************************************
   * Recursively apply SMT solver.
   * @param the folder containing the tasks (and other folders).
   ***************************************************************************/
  private void applySMT(TaskFolder folder)
  {
    top.startExecution(false, new Runnable()
    {
      public void run() 
      {
        if(SMT.enableiterativeTheoremChecking)
          applySMTIteratively(folder);
        else
          applySMTSync(folder);
      }
    });
  }
  
  private void applySMTIteratively(TaskFolder folder)
  {
    Main.resetStopped();    
    List<Task> tasks=getTasks(folder);  
    List<TheoremParamDefinition> thms=new ArrayList<TheoremParamDefinition>();
    List<Consumer<Boolean>> handlers = new ArrayList<Consumer<Boolean>>();
    for(Task t:tasks)
    {
      FunctionSymbol fun = t.getFunction();
      Declaration def = fun.getDefinition();
      thms.add((TheoremParamDefinition)def);  
      handlers.add((Boolean result)->
      {
        top.updateView(new Runnable() {
          public void run() {
            TreeItem item = find(tree.getItems(), t);
            tree.deselect(item);
            if (result)
            {
              item.setForeground(taskColorClosed);
              item.setImage(top.getTaskClosedImage());
            }
          }});
      });
    }
    if(thms.size()==0)
      return;
    List<Declaration> decs=getDeclarations(folder);
    if(decs==null)
      return;
    List<Boolean> results=applySMTCoreIterative(thms,decs,handlers);
    top.updateView(new Runnable() {
      public void run() {
        ListIterator<Boolean> it=results.listIterator();
        for(Task t:tasks)
        {
          if(!it.hasNext())
            return;
          TreeItem item = find(tree.getItems(), t);
          tree.deselect(item);
          if (it.next())
          {
            item.setForeground(taskColorClosed);
            item.setImage(top.getTaskClosedImage());
          }
        }
      }});
  }  

  private List<Task> getTasks(TaskFolder folder)
  {
    List<Task> tasks=new ArrayList<Task>();
    for(Task t:folder.getTasks())
    {
      FunctionSymbol fun = t.getFunction();
      if(fun!=null)
      {
        Declaration def = fun.getDefinition();
        if(def instanceof TheoremParamDefinition)
          tasks.add(t);
      }
    }
    for(TaskFolder tF:folder.getFolders())
    {
      tasks.addAll(getTasks(tF));
    }
    return tasks;
  }


  private List<Declaration> getDeclarations(TaskFolder folder)
  {
    for(Task t:folder.getTasks())
    {
      List<Declaration> declarations = t.getDeclarations();
      if(declarations!=null)
        return declarations;
    }
    for(TaskFolder tF:folder.getFolders())
    {
      List<Declaration> lst=getDeclarations(tF);
      if(lst!=null)
        return lst;
    }
    return null;
  }


  private List<Boolean> SMT_Iterativeresults;

  @SuppressWarnings("deprecation")
  private List<Boolean> applySMTCoreIterative(List<TheoremParamDefinition> theorems,
    List<Declaration> declarations, List<Consumer<Boolean>> handlers)
  {
    top.run(new Runnable() {
      public void run()
      {
        SMT_Iterativeresults = Collections.nCopies(theorems.size(), false);
        smtThread = new Thread(()-> 
        { 
          try
          {
            SMT_Iterativeresults=SMT.checkMultipleTheorems(theorems, declarations, handlers);
          }
          catch(Main.Stopped e) 
          { 
            Main.getOutput().println("ERROR: execution was stopped.");
          }
        });
        smtThread.start();
        try
        {
          while (true)
          {
            try { Thread.sleep(100); } catch(InterruptedException e) { }
            if (!smtThread.isAlive()) break;
            Main.checkStopped();
          }
        }
        catch(Exception e)
        {
          SMT.terminateSolver();
          smtThread.interrupt();
          try { Thread.sleep(500); } catch(InterruptedException e0) { }
          if (smtThread.isAlive()) 
          {
            Main.getOutput().println("SMT thread could not be gracefully interrupted; it is forcefully stopped.");
            smtThread.stop();
          }
          if (e instanceof Main.Stopped)
            Main.getOutput().println("ERROR: execution was stopped.");
        }
      }});
    return SMT_Iterativeresults;
  }
  
  /****************************************************************************
   * Create the event handlers for the tree.
   ***************************************************************************/
  public void createHandlers()
  {
    // set expansion handler
    tree.addTreeListener(new TreeListener() {
      public void treeExpanded(TreeEvent e) {
      }

      public void treeCollapsed(TreeEvent e) {
      }
    });
    
    // set click handlers
    tree.addSelectionListener(new SelectionAdapter() {
      
      // item is double clicked
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        Task task = (Task)item.getData();
        executeTask(task, item);
      }
      
      // item is selected by single click
      public void widgetSelected(SelectionEvent e) 
      {
        TreeItem item = (TreeItem)e.item;
        Task task = (Task)item.getData();
        SourcePosition pos = task.getSourcePosition();
        if (pos == null) pos = fun.ident.getPosition();
        top.markText(pos, true);
        for (SourcePosition pos0 : task.getAuxiliaryPositions())
          top.markText(pos0, false);
      }
    });
    
    // set hover handler
    tree.addMouseTrackListener(new MouseTrackAdapter() {
      public void mouseHover(MouseEvent e) {
        Tree tree = (Tree)e.widget;
        Task task = find(tree.getItems(), e.x, e.y);
        if (task == null) return;
        new TaskTip(top.getShell(), tree.toDisplay(e.x, e.y), task);
      }
      public void mouseExit(MouseEvent e) {
        TaskTip.close();
      }
    });
  }

  /****************************************************************************
   * Find task associated to item in set of items denoted by mouse position.
   * @param items the set of tree items to be recursively searched.
   * @param x horizontal coordinate of mouse position.
   * @param y vertical coordinate of mouse position.
   * @return the denoted task (or null, if none was found)
   ***************************************************************************/
  private Task find(TreeItem[] items, int x, int y)
  {
    for (TreeItem item : items)
    {
      TreeItem[] items0 = item.getItems();
      Task task = find(items0, x, y);
      if (task != null) return task;
      task = (Task)item.getData();
      if (task == null) continue;
      Rectangle b = item.getBounds();
      if (x < b.x || x > b.x + b.width) continue;
      if (y < b.y || y > b.y + b.height) continue;
      return task;
    }
    return null;
  }
  
  /****************************************************************************
   * Display tasks in given root folder.
   * @param root the root folder for display (null clears task tree)
   * @param fun the function to which the tasks belong (null if none)
   ***************************************************************************/
  public void displayTasks(TaskFolder root, FunctionSymbol fun)
  {
    // clear status information
    this.fun = null;
    tree.removeAll();
    
    // process top-level tasks and folders
    if (root != null)
    {
      this.fun = fun;
      Task[] tasks = root.getTasks();
      for (Task t : tasks)
        displayTasks(tree, t);
      TaskFolder[] folders = root.getFolders();
      for (TaskFolder f : folders)
        displayTasks(tree, f);
    }
  }
  
  /***************************************************************************
   * Display task within parent component.
   * @param parent the parent component (type Tree or TreeItem).
   * @param task the task to be displayed.
   **************************************************************************/
  private void displayTasks(Object parent, Task task)
  {
    // create tree node and associate task to it
    TreeItem node = (parent instanceof Tree) ?
        new TreeItem((Tree)parent,SWT.NULL) :
        new TreeItem((TreeItem)parent, SWT.NULL);
    node.setExpanded(true);
    
    // set tree node content
    node.setText(task.getName());
    node.setData(task);

    // if task is atomic, we are done
    if (!(task instanceof TaskFolder)) 
    {  
      if (task.getFunction() != null)
      {
        node.setForeground(taskColorOpen);
        node.setImage(top.getTaskOpenImage());
      }
      else
        node.setForeground(noTaskColor);
      return;
    }
    
    // otherwise, process tasks and folders recursively
    TaskFolder folder = (TaskFolder)task;
    Task[] tasks = folder.getTasks();
    for (Task t : tasks)
      displayTasks(node, t);
    TaskFolder[] folders = folder.getFolders();
    for (Task t : folders)
      displayTasks(node, t);
  }
  
  /****************************************************************************
   * Popup information on a task
   ***************************************************************************/
  private final static class TaskTip 
  {
    private Shell shell;
    private Label label;
    
    // only one instance of task tip at a time
    private static TaskTip tip = null;
    
    /**************************************************************************
     * Create popup window displaying information on given task.
     * @param parent the parent window.
     * @param pos the position of the window.
     * @param task the task.
     *************************************************************************/
    public TaskTip(Shell parent, Point pos, Task task)
    {         
      // close any previous tip
      close();
      tip = null;
      
      // determine text 
      String text = task.getDescription();
      if (text == null) return;
      
      // cthis will be the new tipi
      tip = this;
      
      // create window
      shell = new Shell(parent, SWT.ON_TOP | SWT.TOOL); 
      GridLayout gridLayout = new GridLayout();
      gridLayout.numColumns = 1;
      gridLayout.marginWidth = 2;
      gridLayout.marginHeight = 2;
      shell.setLayout(gridLayout);
      Color color = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_BACKGROUND);
      Color tcolor = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_FOREGROUND);
      shell.setBackground(color);
      
      // set label
      label = new Label(shell, SWT.NONE);
      label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL
          | GridData.VERTICAL_ALIGN_CENTER));
      label.setText(text);
      label.setBackground(color);
      label.setForeground(tcolor);
      
      // add listener
      shell.addMouseTrackListener(new MouseTrackAdapter() { 
        public void mouseEnter(MouseEvent e)
        {
          close();
        }
      });
      
      // open window
      shell.pack();
      MainSWT.setLocation(shell, pos.x+10, pos.y+10);
      shell.open();
    }
    
    /*************************************************************************
     * Close tip instance.
     ************************************************************************/
    public static void close()
    {
      if (tip != null) tip.shell.close();
      tip = null;
    }
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------