// ---------------------------------------------------------------------------
// Inline.java
// Inline function applications by definitions or contracts.
// $Id: Inline.java,v 1.2 2021/05/19 17:28:02 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2017-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.tasks;

import riscal.syntax.*;
import riscal.syntax.AST.*;
import riscal.syntax.AST.Expression.*;
import riscal.types.Environment.Symbol.*;

import java.util.*;

// code reused from UseContract.java, but adapted such that the result can be
// used without further type-checking (already annotated with symbols)
public class Inline extends ASTCloner
{
  /***************************************************************************
   * Inline function applications by definitions or contracts.
   * @param exp the expression to be transformed.
   * @param fun the function in which this expression occurs (may be null)
   * @param callers the caller map.
   * @return the transformed expression.
   **************************************************************************/
  public static Expression execute(Expression exp, 
    FunctionSymbol fun, Recursion.CallMap callers)
  {
    Expression exp0 = (Expression)exp.accept(new Inline(fun, callers));
    return exp0;
  }

  // private FunctionSymbol current;    // the current function (may be null)
  private Recursion.CallMap callers; // the call map of the current specification

  private Inline(FunctionSymbol current, Recursion.CallMap callers)
  {
    // this.current = current;
    this.callers = callers;
  }
  
  public Identifier visit(Identifier id)
  {
    // do not clone identifier, preserve its symbol
    return id;
  }
  
  public Expression visit(ApplicationExp exp)
  {
    List<Expression> exps0 = new ArrayList<Expression>();
    for (Expression e : exp.exps) 
    { 
      Expression e0 = annotate(e, (Expression)e.accept(this));
      exps0.add(e0);
    }
    Expression exp0 = transform(exp.ident, exps0);
    return annotate(exp, exp0);
  }
  
  /***************************************************************************
   * Transform function application.
   * @param fident the identifier of the function.
   * @param args the arguments of the function.
   * @return the transformed expression.
   **************************************************************************/
  private Expression transform(Identifier fident, List<Expression> args)
  {
    FunctionSymbol fun = (FunctionSymbol)fident.getSymbol();
    Declaration def = fun.getDefinition();
    OpInfo info = OpInfo.getInfo(def);
    List<Identifier> params = new ArrayList<Identifier>();
    for (Parameter p : info.param) params.add(p.ident);
    
    // replace application of operation by its contract if
    // - the operation is a procedure OR
    // - OLD: the current function is called recursively by this operation OR
    // - NEW: the operation is (directly or indirectly recursive)
    // - the function is annotated as "modular"
    boolean doit = 
       def instanceof Declaration.ProcedureDefinition || 
       // (current != null && callers.calls(fun, current));
       callers.calls(fun, fun);
    if (!doit)
    {
      for (FunctionSpec s : info.spec)
      {
        doit = s instanceof FunctionSpec.ContractSpec;
        if (doit) break;
      }
    }
    if (doit) 
    {
      // replace application by choice of value satisfying the operation's postcondition
      Expression postFormula = info.postFormula;
      if (postFormula == null) 
      {
        postFormula = new Expression.TrueExp();
        postFormula.setType(new Type.BoolType());
      }
      List<QuantifiedVariableCore> qv0 = new ArrayList<QuantifiedVariableCore>();
      ValueSymbol result = fun.getResult();
      qv0.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(result.ident, fun.type, postFormula));
      QuantifiedVariable qvar = new QuantifiedVariable(qv0, null);
      Expression choice = new Expression.ChooseExp(qvar);
      choice.setType(fun.type);
      return CloneIds.bind(params, args, choice);
    }
    
    // replace application by (also inlined) and instantiated function body
    Expression exp = execute(info.bodyexp, fun, callers);
    return CloneIds.bind(params, args, exp);
  }
    
  /**************************************************************************
   * Annotate clone with information from original expression.
   * @param original the original expression
   * @param clone the clone
   * @return the clone
   *************************************************************************/
  private static Expression annotate(Expression original, Expression clone)
  {
    clone.setType(original.getType());
    clone.setPosition(original.getPosition());
    return clone;
  }
  
  public Expression visit(AndExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ArrayBuilderExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MapBuilderExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MapSelectionExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MapUpdateExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(AssertExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(BigUnionExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(BigIntersectExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ChooseExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ChooseInElseExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ChooseInExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(DividesExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(EmptySetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(EnumeratedSetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(CartesianExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(EqualsExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(EquivExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ExistsExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(FalseExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ForallExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(GreaterEqualExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(GreaterExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(HeightExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(IdentifierExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(IfThenElseExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ImpliesExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(InSetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NotInSetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(IntersectExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(IntervalExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MatchExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NumberLiteralExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(LessEqualExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(LessExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(LetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(LetParExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MinusExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NegationExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(FactorialExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NotEqualsExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NotExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NumberExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(OrExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PlusExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PlusExpMult ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PowerExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PowersetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(Powerset1Exp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(Powerset2Exp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PrintExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PrintInExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(PrintTraceInExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(CheckExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RecApplicationExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RecIdentifierExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RecordExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RecordSelectionExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RecordUpdateExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(RemainderExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(SetBuilderExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(SumExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(ProductExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MinExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(MaxExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(SetSizeExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(SubsetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(NotSubsetExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TimesExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TimesExpMult ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TrueExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TupleExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TupleSelectionExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(TupleUpdateExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(UnionExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(UnitExp ast) { return annotate(ast, super.visit(ast)); }
  public Expression visit(WithoutExp ast) { return annotate(ast, super.visit(ast)); }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
